/*++
Module Name:
    driver.cpp

Abstract:
    This file contains the driver entry points and callbacks.

Environment:
    Kernel-mode Driver Framework
--*/

#include "driver.h"
#include "driver.tmh"

#ifdef ALLOC_PRAGMA
#pragma alloc_text (INIT, DriverEntry)
#pragma alloc_text (PAGE, AIOWDM4EvtDriverContextCleanup)
#endif

PDEVICE_OBJECT pCentralDevice = NULL;

NTSTATUS
DriverEntry(
    _In_ PDRIVER_OBJECT  DriverObject,
    _In_ PUNICODE_STRING RegistryPath
)
/*++
Routine Description:
    DriverEntry initializes the driver and is the first routine called by the
    system after the driver is loaded. DriverEntry specifies the other entry
    points in the function driver, such as EvtDevice and DriverUnload.

Parameters Description:
    DriverObject - represents the instance of the function driver that is loaded
    into memory. DriverEntry must initialize members of DriverObject before it
    returns to the caller. DriverObject is allocated by the system before the
    driver is loaded, and it is released by the system after the system unloads
    the function driver from memory.

    RegistryPath - represents the driver-specific path in the Registry.
    The function driver can use the path to store driver related data between
    reboots. The path does not store hardware-instance-specific data.

Return Value:
    STATUS_SUCCESS if successful,
    STATUS_UNSUCCESSFUL otherwise.
--*/
{
    WDF_DRIVER_CONFIG config;
    NTSTATUS status;
    WDF_OBJECT_ATTRIBUTES attributes;

    // Initialize WPP Tracing
    WPP_INIT_TRACING(DriverObject, RegistryPath);

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry");

    // Register a cleanup callback so that we can call WPP_CLEANUP when
    // the framework driver object is deleted during driver unload.
    WDF_OBJECT_ATTRIBUTES_INIT(&attributes);
    attributes.EvtCleanupCallback = AIOWDM4EvtDriverContextCleanup;

    WDF_DRIVER_CONFIG_INIT(&config, AIOWDM4EvtDeviceAdd);

    status = WdfDriverCreate(DriverObject,
        RegistryPath,
        &attributes,
        &config,
        WDF_NO_HANDLE
    );
    if (!NT_SUCCESS(status)) {
        TraceEvents(TRACE_LEVEL_ERROR, TRACE_DRIVER, "WdfDriverCreate failed %!STATUS!", status);
        WPP_CLEANUP(DriverObject);
        goto funcexit;
    }

    // Create central device for global commands.
    UNICODE_STRING NtDeviceName;
    UNICODE_STRING Win32DeviceName;
    RtlInitUnicodeString(&NtDeviceName, DEVICE_DEVICE_NAME);
    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    IoCreateSymbolicLink(&Win32DeviceName, &NtDeviceName);

    if (FALSE) // --J2H Debug
        if (pCentralDevice == NULL)
        {
            pCentralDevice = AIOWDMCreateCentralDevice(DriverObject);
            if (pCentralDevice == NULL)
                TraceEvents(TRACE_LEVEL_ERROR, TRACE_DRIVER, "AIOWDMCreateCentralDevice failed");
        }

funcexit:
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit status: %!STATUS!", status);

    return status;
}

VOID
AIOWDM4EvtDriverContextCleanup(
    _In_ WDFOBJECT DriverObject
)
/*++
Routine Description:
    Free all the resources allocated in DriverEntry.

Arguments:
    DriverObject - handle to a WDF Driver object.

Return Value:
    VOID.
--*/
{
    UNREFERENCED_PARAMETER(DriverObject);

    PAGED_CODE();
    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry");

    UNICODE_STRING Win32DeviceName;

    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    IoDeleteSymbolicLink(&Win32DeviceName);

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit");

    //
    // Stop WPP Tracing
    //
    WPP_CLEANUP(WdfDriverWdmGetDriverObject((WDFDRIVER)DriverObject));
}


PDEVICE_OBJECT AIOWDMCreateCentralDevice(IN PDRIVER_OBJECT DriverObject)
{
    NTSTATUS Status;
    UNICODE_STRING NtDeviceName;
    UNICODE_STRING Win32DeviceName;
    PDEVICE_OBJECT deviceObject;

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> %!FUNC! Entry ");

    RtlInitUnicodeString(&NtDeviceName, DEVICE_DEVICE_NAME);

    Status = IoCreateDevice(
        DriverObject,
        sizeof(DEVICE_CONTEXT),
        &NtDeviceName,
        FILE_DEVICE_UNKNOWN,
        0,
        FALSE, // Not exclusive
        &deviceObject
    );

    if (!NT_SUCCESS(Status))
    {
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> IoCreateDevice failed status %d\n", Status);
        deviceObject = NULL;
        goto funcexit;
    }

    RtlInitUnicodeString(&Win32DeviceName, AIOWDM_DEVICE_NAME);
    Status = IoCreateSymbolicLink(&Win32DeviceName, &NtDeviceName);
    if (!NT_SUCCESS(Status)) // If we we couldn't create the link then abort installation.
    {
        IoDeleteDevice(deviceObject);
        TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "--> IoCreateSymbolicLink failed status %d\n", Status);
        deviceObject = NULL;
        goto funcexit;
    }

    PDEVICE_CONTEXT deviceContext = (PDEVICE_CONTEXT)deviceObject->DeviceExtension;
    AIOWDM4InitDefaultDeviceContext(deviceContext);

    TraceEvents(TRACE_LEVEL_INFORMATION, TRACE_DRIVER, "<-- %!FUNC! Exit ");

funcexit:
    return deviceObject;
}