/*++
Module Name:
    driver.h

Abstract:
    This file contains the driver definitions.

Environment:
    Kernel-mode Driver Framework
--*/

#pragma once

#include <ntddk.h>
#include <wdf.h>
#include <initguid.h>
#include "trace.h"

EXTERN_C_START

#ifdef _WIN64
typedef ULONGLONG UNATIVE;
#else
typedef ULONG UNATIVE;
#endif

#define AIOWDM4_ACK_INTERRUPT(_adapter, _value) { \
   *_value = READ_PORT_ULONG(PULONG(_adapter->MainBar.Base + COSIRQLatchOffset)); \
   WRITE_PORT_ULONG(PULONG(_adapter->MainBar.Base + COSIRQLatchOffset), *_value); }


typedef struct _IOACTION
{
    char BarID;
    char Operation;
    UCHAR Offset;
    UCHAR Data;
} IOACTION, * PIOACTION;

typedef enum _TBarType { btNone, btMemory, btIO, btMappedIO } TBarType;
typedef struct _TBarData
{
    TBarType BarType;
    PUCHAR Base;
    ULONG Size;
    PHYSICAL_ADDRESS UnmappedBase;
} TBarData, * TPBarData;

typedef struct _TCOSScheme
{
    UCHAR StartOff, Size, Stride;
} TCOSScheme, * TPCOSScheme;

typedef union
{
    UCHAR Data8;
    USHORT Data16;
    ULONG Data32;
}DATA;

typedef struct _TMEMOP
{
    ULONG Address;
    USHORT Padding0;
    UCHAR Flags;
    UCHAR Size;
    DATA Data;
} TMEMOP, * PMEMOP;

typedef struct _TPORTOP
{
    USHORT Address;
    UCHAR Flags;
    UCHAR Size;
    DATA Data;
} TPORTOP, * PPORTOP;

typedef enum
{
	STOPPED,								// device stopped
	WORKING,								// started and working
	PENDINGSTOP,							// stop pending
	PENDINGREMOVE,							// remove pending
	SURPRISEREMOVED,						// removed by surprise
	REMOVED,								// removed
} DEVSTATE;

typedef struct _TCARDINFO
{
	ULONG DeviceID;
	ULONG BaseAddress;
} TCARDINFO, * PCARDINFO;


///////////////////////////////////////////////////////////////////////////////
// Device extension structure

// The "device context" performs the same job as a "WDM device extension" in the driver frameworks
typedef struct _DEVICE_CONTEXT
{
	WDFDEVICE WdfDeviceObject;			// device object this extension belongs to
	WDFDEVICE LowerDeviceObject;		// next lower driver in same stack
	PDEVICE_OBJECT Pdo;						// the PDO
	IO_REMOVE_LOCK RemoveLock;				// removal control locking structure
	UNICODE_STRING ifname;					// interface name
	DEVSTATE state;							// current state of device
	DEVSTATE prevstate;						// state prior to removal query
	DEVICE_POWER_STATE devpower;			// current device power state
	SYSTEM_POWER_STATE syspower;			// current system power state
	DEVICE_POWER_STATE PerfBoundary;		// context restore very expensive from >= this power state
	POWER_SEQUENCE oldseq;					// old power sequence numbers
	DEVICE_CAPABILITIES devcaps;			// copy of most recent device capabilities
	LONG handles;							// # open handles
	WDFINTERRUPT InterruptObject;			// address of interrupt object

	BOOLEAN bRealDevice; //false if central device(just for global commands)
	ULONG DeviceID;							// The card's PCI(e) device ID or ISA virtual device ID
	TBarData StatusBar, MainBar, AuxBar;
	IOACTION IRQIsEnabled;					// how to determine if the card-level IRQ is enabled
	IOACTION IRQIAm;						// how to find out if the device is generating an IRQ
	IOACTION IRQDisable;					// how to disable IRQs
	IOACTION IRQClear;						// how to clear IRQs
	IOACTION IRQEnable;						// how to enable IRQs
	ULONG InterruptOccurred;				// --J2H ToDo: change to per-handle flag for multi-process waitForIRQ
	ULONG InterruptRequested;
	WDFREQUEST PendingIrp;
	BOOLEAN HasPendingIrp;
	ULONG COSSize;
	TCOSScheme COSScheme;
	WDFQUEUE IRQQueue;

	WDFSPINLOCK RcvLock;
	WDFSPINLOCK Lock;
	WDFSPINLOCK SendLock;
	BUS_INTERFACE_STANDARD  BusInterface;
} DEVICE_CONTEXT, * PDEVICE_CONTEXT;
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(DEVICE_CONTEXT, AIOWDM4DeviceGetContext)

// This is the context that can be placed per queue
// and contains per queue information.
typedef struct _QUEUE_CONTEXT {
	PDEVICE_CONTEXT deviceContext;
	PPORTOP Data;
} QUEUE_CONTEXT, * PQUEUE_CONTEXT;
WDF_DECLARE_CONTEXT_TYPE_WITH_NAME(QUEUE_CONTEXT, QueueGetContext)

#define maskIAMIRQ 0x80
#define maskDisableIRQ 0x00
#define maskEnableIRQ 0x09
#define maskInterrupting (maskIAMIRQ | maskEnableIRQ)

#define IRQEnableOffset 0x69
#define IRQStatusOffset 0x0F
#define COSIRQLatchOffset 0x08


#define DEVICE_DEVICE_NAME L"\\Device\\AIOWDMCoreDevice"
#define AIOWDM_DEVICE_NAME L"\\DosDevices\\AIOWDMCore"

#define FILE_DEVICE_KRNLDRVR 0x80ff

#define IOCTL_QUERY_INFO     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x800, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_QUERY_BAR_BASE CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x801, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_IRQ_WAIT_BEGIN CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x810, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_IRQ_WAIT_ABORT CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x812, METHOD_BUFFERED, FILE_ANY_ACCESS )
#define IOCTL_IRQ_POLL       CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x814, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_SET_COS_SCHEME CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x816, METHOD_BUFFERED, FILE_READ_ACCESS)
#define IOCTL_FORGET_IRP     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x880, METHOD_BUFFERED, FILE_WRITE_ACCESS)

#define IOCTL_PORT_IN        CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x820, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_PORT_INSTR     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x821, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_PORT_OUT       CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x822, METHOD_BUFFERED, FILE_WRITE_ACCESS)
//#define IOCTL_PORT_OUTSTR    CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x823, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define IOCTL_MEM_IN         CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x824, METHOD_BUFFERED, FILE_READ_ACCESS )
#define IOCTL_MEM_OUT        CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x826, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define IOCTL_PORTMB_OUT     CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x82A, METHOD_BUFFERED, FILE_WRITE_ACCESS)
#define IOCTL_GET_BENCH      CTL_CODE(FILE_DEVICE_KRNLDRVR, 0x830, METHOD_BUFFERED, FILE_READ_ACCESS )

void AIOWDM4InitDefaultDeviceContext(PDEVICE_CONTEXT deviceContext);

EVT_WDF_INTERRUPT_ISR AIOWDM4EvtInterruptIsr;
EVT_WDF_INTERRUPT_DPC AIOWDM4EvtInterruptDpc;
EVT_WDF_INTERRUPT_ENABLE AIOWDM4EvtInterruptEnable;
EVT_WDF_INTERRUPT_DISABLE AIOWDM4EvtInterruptDisable;

UCHAR InBar8(const TBarData Bar, ULONG Offset);
USHORT InBar16(const TBarData Bar, ULONG Offset);
ULONG InBar32(const TBarData Bar, ULONG Offset);

void OutBar8(const TBarData Bar, ULONG Offset, UCHAR Data);
void OutBar16(const TBarData Bar, ULONG Offset, USHORT Data);
void OutBar32(const TBarData Bar, ULONG Offset, ULONG Data);
void ForgetIRP(PIRP IRP);




NTSTATUS AIOWDM4IOCTLQueueInitialize(_In_ WDFDEVICE Device);

EVT_WDF_DRIVER_DEVICE_ADD AIOWDM4EvtDeviceAdd;
NTSTATUS AIOWDM4CreateIRQQueue(_In_ WDFDEVICE Device);
PDEVICE_OBJECT AIOWDMCreateCentralDevice(IN PDRIVER_OBJECT DriverObject);

//
// Function to initialize the device and its callbacks
//
NTSTATUS AIOWDM4CreateDevice(_In_ PWDFDEVICE_INIT DeviceInit);
NTSTATUS AIOWDM4GetDeviceInformation(IN OUT PDEVICE_CONTEXT deviceContext);
NTSTATUS
AIOWDM4MapHWResources(
	_In_ OUT PDEVICE_CONTEXT deviceContext,
	_In_ WDFCMRESLIST  ResourcesRaw,
	_In_ WDFCMRESLIST  ResourcesTranslated
);
NTSTATUS
AIOWDM4UnmapHWResources(
	_In_ OUT PDEVICE_CONTEXT deviceContext
);
NTSTATUS
AIOWDM4EvtPrepareHardware(
	WDFDEVICE Device,
	WDFCMRESLIST ResourcesRaw,
	WDFCMRESLIST ResourcesTranslated
);

VOID InCOSData(PDEVICE_CONTEXT pdx, PUCHAR pDst, size_t COSSize);


DRIVER_INITIALIZE DriverEntry;
EVT_WDF_OBJECT_CONTEXT_CLEANUP AIOWDM4EvtDriverContextCleanup;

// Events from the IoQueue object

EVT_WDF_IO_QUEUE_IO_DEVICE_CONTROL AIOWDM4EvtIoDeviceControl;
EVT_WDF_IO_QUEUE_IO_STOP AIOWDM4EvtIoStop;

EXTERN_C_END